import can
import logging
import os

from time import time
from time import time_ns

import cand.config as cfg
from cand.util import get_msg_by_name
from cand.serialization import deserialize, serialize

MAX_BUFFER_SIZE = 25
MAX_BUFFER_WAIT_TIME_SECONDS = 0.001
QUEUE_POP_TIMEOUT = 0.0001


def talker():
    log = logging.getLogger("talker")
    log.info(f"-- Started cand talker with PID {os.getpid()} --")

    time_at_last_buffer_warning = 0
    buffer_warning_suppress_count = 0

    time_at_flush = time()
    p = cfg.rdb.pipeline()

    while True:
        # Check if we need to flush the buffer
        if len(p) >= MAX_BUFFER_SIZE:
            p.execute()
            time_at_flush = time()

        elif len(p) > 0 and time() - time_at_flush > MAX_BUFFER_WAIT_TIME_SECONDS:
            p.execute()
            time_at_flush = time()

        [_, serial_msgs] = cfg.rdb.blmpop(
            QUEUE_POP_TIMEOUT,
            1,
            "queue:cansend",
            direction="LEFT",
            count=MAX_BUFFER_SIZE + 1,
        )

        if serial_msgs is None:
            continue

        if len(serial_msgs) > MAX_BUFFER_SIZE:
            if time() - time_at_last_buffer_warning > 1:
                log.critical(
                    f"Buffer is saturated ({len(serial_msgs)})! cand may not be keeping up with sent messsages."
                    + f" (suppressed {buffer_warning_suppress_count} times)"
                )
                buffer_warning_suppress_count = 0
                time_at_last_buffer_warning = time()
            else:
                buffer_warning_suppress_count += 1

        for serial_msg in serial_msgs:
            name = None
            data = None

            try:
                name, data = deserialize(serial_msg)
                msg = get_msg_by_name(name)
                packed_data = msg.encode(data)
            except Exception as e:
                log.warning(f"Failed to prepare message: {e}")
                continue

            bus_msg = can.Message(
                arbitration_id=msg.frame_id, data=packed_data, is_extended_id=False
            )

            cfg.bus.send(bus_msg)

            # python-can does not loopback messages we send by design,
            # to get around this limitation we'll need to manually set
            # the sent message in our redis database
            p.set(name, serialize((time_ns(), data)))



import time

from cand.client import Bus


def main():
    bus = Bus()

    iterations = 10000
    i = 0
    start_time = time.time()
    while i - 1 < 100000000:
        bus.send("dbwNode_Status", {"SystemStatus": "ESTOP", "Counter": 141})
        # time.sleep(0.00000001)

        if not i % iterations:
            print(
                f"Messages sent / sec: {int(iterations / (time.time() - start_time))}, total: {i:,}"
            )
            start_time = time.time()

        i += 1


if __name__ == "__main__":
    main()



# Global configuration

import can
import cantools
import redis


def init(_bus: can.BusABC, _dbc: cantools.database, _rdb: redis.Redis):
    global bus, dbc, rdb

    bus = _bus
    dbc = _dbc
    rdb = _rdb



import msgpack

from cantools.database.can.signal import NamedSignalValue


def _deserialize_NamedSignalValue(obj):
    if "__!NamedSignalValue__" in obj:
        obj = NamedSignalValue(
            value=obj["_value"], name=obj["_name"], comments=obj["_comments"]
        )
    return obj


def _serialize_NamedSignalValue(obj):
    if isinstance(obj, NamedSignalValue):
        return {
            "__!NamedSignalValue__": True,
            "_name": obj._name,
            "_value": obj._value,
            "_comments": obj._comments,
        }
    return obj


def serialize(obj):
    return msgpack.packb(obj, default=_serialize_NamedSignalValue)


def deserialize(obj):
    return msgpack.unpackb(obj, object_hook=_deserialize_NamedSignalValue)



import logging
import redis

from time import time_ns

from cand.serialization import serialize, deserialize

REDIS_SOCKET_CONNECT_TIMEOUT = 1


class Bus:
    def __init__(self, redis_host: str = "localhost", redis_port: str = "6379"):

        self._log = logging.getLogger("candclient")

        try:
            self._rdb = redis.Redis(
                host=redis_host,
                port=redis_port,
                socket_connect_timeout=REDIS_SOCKET_CONNECT_TIMEOUT,
            )
        except redis.ConnectionError as e:
            self._log.error(f"Failed to connect to Redis: {e}")
            exit(-1)

    def get(self, name: str) -> tuple[int, dict]:
        try:
            data = self._rdb.get(name)

            if data is None:
                self._log.debug(f"Tried to get message '{name}' and it was unavailable")
                return None

            data = deserialize(data)

            return tuple(data)

        except Exception as e:
            self._log.error(f"Error getting message {name}: {e}")
            raise e

    def get_data(self, name: str) -> dict:
        data = self.get(name)

        if data is not None:
            return data[1]

    def get_time(self, name: str) -> int:
        data = self.get(name)

        if data is not None:
            return data[0]

    def get_time_delta(self, name: str) -> int:
        data = self.get(name)

        if data is not None:
            return time_ns() - data[0]

    def send(self, name: str, data: dict) -> None:
        try:
            data = serialize((name, data))

            self._rdb.rpush("queue:cansend", data)

        except Exception as e:
            self._log.error(f"Error sending message {name} with data {data}: {e}")
            raise e



import cantools

from functools import cache

import cand.config as cfg


@cache
def get_msg_by_name(name: str) -> cantools.database.can.message.Message:
    return cfg.dbc.get_message_by_name(name)


@cache
def get_msg_by_frameid(name: str) -> cantools.database.can.message.Message:
    try:
        return cfg.dbc.get_message_by_frame_id(name)
    except KeyError:
        return None



#!/usr/bin/env python3

import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="opencan-cand",
    version="0.0.3",
    author="OpenCAN",
    author_email="info@opencan.org",
    description="CAN Service Daemon",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/opencan/cand",
    packages=["cand"],
    install_requires=[
        "cantools>=37.0.7",
        "coloredlogs>=15.0.1",
        "msgpack>=1.0.3",
        "python-can>=3.3.4",
        "redis>=4.2.2",
    ],
    python_requires=">=3.8",
    scripts=["cand/cand"],
)



# Welcome to cand.

import logging
import os

from time import time
from time import time_ns

import cand.config as cfg

from cand.util import get_msg_by_frameid
from cand.serialization import serialize

MAX_BUFFER_WAIT_TIME_SECONDS = 0.001
MAX_BUFFER_SIZE = 10
CAN_RECV_TIMEOUT_SECONDS = 0.005


def cand():
    log = logging.getLogger("listener")
    log.info(f"-- Started cand listener with PID {os.getpid()} --")

    p = cfg.rdb.pipeline()  # Redis buffer ("pipeline")
    time_at_flush = time()

    log.info("Starting listen loop.")
    while True:
        # Check if we need to flush the buffer
        if len(p) >= MAX_BUFFER_SIZE:
            p.execute()
            time_at_flush = time()

        elif len(p) > 0 and time() - time_at_flush > MAX_BUFFER_WAIT_TIME_SECONDS:
            p.execute()
            time_at_flush = time()

        # Poll the bus for a message.
        #
        # Note: Right now, we're just waiting for the timeout period.
        # If cand gains more functionality where it has to service things
        # other than the bus in this loop, it would be smart to use the
        # asyncio recv and do other stuff for a little while.
        #
        # This timeout is also currently the only thing rate-limiting this
        # loop. We don't want to sleep!
        raw_msg = cfg.bus.recv(timeout=CAN_RECV_TIMEOUT_SECONDS)
        if raw_msg is None:
            continue

        # Find the dbc message
        frame_id = raw_msg.arbitration_id

        msg = get_msg_by_frameid(frame_id)
        if msg is None:  # unrecognized frame ID
            log.debug(f"Unrecognized message: {' '.join(str(raw_msg).split())}")
            continue

        # Serialize and post the message data
        data = (time_ns(), msg.decode(raw_msg.data))
        serialized_data = serialize(data)

        p.set(msg.name, serialized_data)  # commit to Redis buffer!
